import os, sys
parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parent_dir)

from janome.dic import SystemDictionary
from . import entries1, entries2, entries3, entries4, entries5, entries6, entries7, entries8, entries9, entries10
from . import connections1, connections2
from . import chardef, unknowns

def entries():
  entries = entries1.DATA.copy()
  entries.update(entries2.DATA)
  entries.update(entries3.DATA)
  entries.update(entries4.DATA)
  entries.update(entries5.DATA)
  entries.update(entries6.DATA)
  entries.update(entries7.DATA)
  entries.update(entries8.DATA)
  entries.update(entries9.DATA)
  entries.update(entries10.DATA)
  return entries

def connections():
  connections = list(connections1.DATA)
  connections.extend(connections2.DATA)
  return connections

SYS_DIC = SystemDictionary(entries(), connections(), chardef.DATA, unknowns.DATA)
